import json
import ssl
from time import time

import requests
from requests.exceptions import ReadTimeout
from websocket import create_connection
from websocket._exceptions import WebSocketTimeoutException

from .sense_api import *
from .sense_exceptions import *


class Senseable(SenseableBase):
    def __init__(
        self,
        username=None,
        password=None,
        api_timeout=API_TIMEOUT,
        wss_timeout=WSS_TIMEOUT,
        ssl_verify=True,
        ssl_cafile="",
        device_id=None,
    ):
        """Init the Senseable object."""

        # Create session
        self.s = requests.session()
        self.set_ssl_context(ssl_verify, ssl_cafile)

        SenseableBase.__init__(
            self,
            username=username,
            password=password,
            api_timeout=api_timeout,
            wss_timeout=wss_timeout,
            ssl_verify=ssl_verify,
            ssl_cafile=ssl_cafile,
            device_id=device_id,
        )

    def set_ssl_context(self, ssl_verify, ssl_cafile):
        """Create or set the SSL context. Use custom ssl verification, if specified."""
        if not ssl_verify:
            self.s.verify = False
        elif ssl_cafile:
            self.s.verify = ssl_cafile

    def authenticate(self, username, password, ssl_verify=True, ssl_cafile=""):
        """Authenticate with username (email) and password. Optionally set SSL context as well.
        This or `load_auth` must be called once at the start of the session."""
        auth_data = {"email": username, "password": password}

        # Get auth token
        try:
            resp = self.s.post(API_URL + "authenticate", auth_data, headers=self.headers, timeout=self.api_timeout)
        except Exception as e:
            raise Exception("Connection failure: %s" % e)

        # check MFA code required
        if resp.status_code == 401:
            data = resp.json()
            if "mfa_token" in data:
                self._mfa_token = data["mfa_token"]
                raise SenseMFARequiredException(data["error_reason"])

        # check for 200 return
        if resp.status_code != 200:
            raise SenseAuthenticationException(
                "Please check username and password. API Return Code: %s" % resp.status_code
            )

        data = resp.json()
        self._set_auth_data(data)
        self.set_monitor_id(data["monitors"][0]["id"])

    def validate_mfa(self, code):
        """Validate a multi-factor authentication code after authenticate raised SenseMFARequiredException.
        Authentication process is completed if code is valid."""
        mfa_data = {
            "totp": code,
            "mfa_token": self._mfa_token,
            "client_time:": datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
        }
        # Get auth token
        try:
            resp = self.s.post(API_URL + "authenticate/mfa", mfa_data, headers=self.headers, timeout=self.api_timeout)
        except Exception as e:
            raise Exception("Connection failure: %s" % e)

        # check for 200 return
        if resp.status_code != 200:
            raise SenseAuthenticationException(
                "Please check username and password. API Return Code: %s" % resp.status_code
            )

        data = resp.json()
        self._set_auth_data(data)
        self.set_monitor_id(data["monitors"][0]["id"])

    def renew_auth(self):
        renew_data = {
            "user_id": self.sense_user_id,
            "refresh_token": self.refresh_token,
        }

        # Get auth token
        try:
            resp = self.s.post(API_URL + "renew", renew_data, headers=self.headers, timeout=self.api_timeout)
        except Exception as e:
            raise Exception("Connection failure: %s" % e)

        # check for 200 return
        if resp.status_code != 200:
            raise SenseAuthenticationException(
                "Please check username and password. API Return Code: %s" % resp.status_code
            )

        self._set_auth_data(resp.json())

    def logout(self):
        try:
            resp = self.s.get(API_URL + "logout", timeout=self.api_timeout)
        except Exception as e:
            raise Exception("Connection failure: %s" % e)

        # check for 200 return
        if resp.status_code != 200:
            raise SenseAPIException("API Return Code: %s", resp.status_code)

    def update_realtime(self, retry=True):
        """Update the realtime data (device status and current power)."""
        # rate limit API calls
        now = time()
        if self._realtime and self.rate_limit and self.last_realtime_call + self.rate_limit > now:
            return self._realtime
        self.last_realtime_call = now
        url = WS_URL % (self.sense_monitor_id, self.sense_access_token)
        try:
            next(self.get_realtime_stream())
        except SenseAuthenticationException as e:
            if retry:
                self.renew_auth()
                self.update_realtime(False)
            else:
                raise e
                

    def get_realtime_stream(self):
        """Reads realtime data from websocket.  Realtime data variable is set and data is
        returned through generator. Continues until loop broken."""
        ws = 0
        url = WS_URL % (self.sense_monitor_id, self.sense_access_token)
        try:
            ws = create_connection(url, timeout=self.wss_timeout, sslopt={"cert_reqs": ssl.CERT_NONE})
            while True:  # hello, features, [updates,] data
                result = json.loads(ws.recv())
                if result.get("type") == "realtime_update":
                    data = result["payload"]
                    self._set_realtime(data)
                    yield data
                elif result.get("type") == "error":
                    data = result["payload"]
                    if not data["authorized"]:
                        raise SenseAuthenticationException("Web Socket Unauthorized")
                    raise SenseWebsocketException(data["error_reason"])
        except WebSocketTimeoutException:
            raise SenseAPITimeoutException("API websocket timed out")
        finally:
            if ws:
                ws.close()

    def _api_call(self, url, payload={}, retry=False):
        """Make a call to the Sense API directly and return the json results."""
        try:
            resp = self.s.get(
                API_URL + url,
                headers=self.headers,
                timeout=self.api_timeout,
                params=payload,
            )

            if not retry and resp.status_code == 401:
                self.renew_auth()
                return self._api_call(url, payload, True)

            # 4xx represents unauthenticated
            if resp.status_code == 401 or resp.status_code == 403 or resp.status_code == 404:
                raise SenseAuthenticationException("API Return Code: %s", resp.status_code)
            return resp.json()
        except ReadTimeout:
            raise SenseAPITimeoutException("API call timed out")

    def get_trend_data(self, scale, dt=None):
        """Update trend data for specified scale from API.
        Optionally set a date to fetch data from."""
        if scale.upper() not in valid_scales:
            raise Exception("%s not a valid scale" % scale)
        if not dt:
            dt = datetime.utcnow()
        self._trend_data[scale] = self._api_call(
            "app/history/trends?monitor_id=%s&scale=%s&start=%s"
            % (self.sense_monitor_id, scale, dt.strftime("%Y-%m-%dT%H:%M:%S"))
        )

    def update_trend_data(self, dt=None):
        """Update trend data of all scales from API.
        Optionally set a date to fetch data from."""
        for scale in valid_scales:
            self.get_trend_data(scale, dt)

    def get_monitor_data(self):
        """Get monitor overview info from API."""
        json = self._api_call("app/monitors/%s/overview" % self.sense_monitor_id)
        if "monitor_overview" in json and "monitor" in json["monitor_overview"]:
            self._monitor = json["monitor_overview"]["monitor"]
        return self._monitor

    def get_discovered_device_names(self):
        """Get list of device names from API."""
        json = self._api_call("app/monitors/%s/devices" % self.sense_monitor_id)
        self._devices = [entry["name"] for entry in json]
        return self._devices

    def get_discovered_device_data(self):
        """Get list of raw device data from API."""
        return self._api_call("monitors/%s/devices" % self.sense_monitor_id)

    def always_on_info(self):
        """Always on info from API - pretty generic similar to the web page."""
        return self._api_call("app/monitors/%s/devices/always_on" % self.sense_monitor_id)

    def get_monitor_info(self):
        """View info on monitor & device detection status from API."""
        return self._api_call("app/monitors/%s/status" % self.sense_monitor_id)

    def get_device_info(self, device_id):
        """Get specific informaton about a device from API."""
        return self._api_call("app/monitors/%s/devices/%s" % (self.sense_monitor_id, device_id))

    def get_all_usage_data(self, payload={"n_items": 30}):
        """Gets usage data by device from API

        Args:
            payload (dict, optional): known params are:
                n_items: the number of items to return
                device_id: limit results to a specific device_id
                prior_to_item:. date in format YYYY-MM-DDTHH:MM:SS.mmmZ
                rollup: ?

                Defaults to {'n_items': 30}.

        Returns:
            dict: usage data
        """
        # lots of info in here to be parsed out
        return self._api_call("users/%s/timeline" % (self.sense_user_id), payload)



import uuid
from datetime import datetime

from .sense_exceptions import *

API_URL = "https://api.sense.com/apiservice/api/v1/"
WS_URL = "wss://clientrt.sense.com/monitors/%s/realtimefeed?access_token=%s"
API_TIMEOUT = 5
WSS_TIMEOUT = 5
RATE_LIMIT = 60

# for the last day, week, month, or year
valid_scales = ["DAY", "WEEK", "MONTH", "YEAR"]


class SenseableBase(object):
    def __init__(
        self,
        username=None,
        password=None,
        api_timeout=API_TIMEOUT,
        wss_timeout=WSS_TIMEOUT,
        ssl_verify=True,
        ssl_cafile="",
        device_id=None,
    ):
        """Initialize SenseableBase object."""

        # Timeout instance variables
        self.api_timeout = api_timeout
        self.wss_timeout = wss_timeout
        self.rate_limit = RATE_LIMIT
        self.last_realtime_call = 0

        self._mfa_token = ""
        self._realtime = {}
        self._devices = []
        self._trend_data = {}
        self._monitor = {}
        for scale in valid_scales:
            self._trend_data[scale] = {}
        self.set_ssl_context(ssl_verify, ssl_cafile)
        if device_id:
            self.device_id = device_id
        else:
            self.device_id = str(uuid.uuid4()).replace("-", "")

        self.headers = {"x-sense-device-id": self.device_id}

        if username and password:
            self.authenticate(username, password)

    def load_auth(self, access_token, user_id, device_id, refresh_token):
        """Load the authentication data from a previous session."""
        self.device_id = device_id
        data = {
            "access_token": access_token,
            "user_id": user_id,
            "refresh_token": refresh_token,
        }
        self._set_auth_data(data)

    def set_monitor_id(self, monitor_id):
        self.sense_monitor_id = monitor_id

    def _set_auth_data(self, data):
        """Set the authentication data for the session."""
        self.sense_access_token = data["access_token"]
        self.sense_user_id = data["user_id"]
        self.refresh_token = data["refresh_token"]

        # create the auth header
        self.headers = {
            "x-sense-device-id": self.device_id,
            "Authorization": "bearer {}".format(self.sense_access_token),
        }

    @property
    def devices(self):
        """List of discovered device names."""
        return self._devices

    def _set_realtime(self, data):
        """Sets the realtime data structure."""
        self._realtime = data

    def get_realtime(self):
        """Return the raw realtime data structure."""
        return self._realtime

    @property
    def active_power(self):
        return self._realtime.get("w", 0)

    @property
    def active_solar_power(self):
        return self._realtime.get("solar_w", 0)

    @property
    def active_voltage(self):
        return self._realtime.get("voltage", [])

    @property
    def active_frequency(self):
        return self._realtime.get("hz", 0)

    @property
    def daily_usage(self):
        return self.get_trend("DAY", "consumption")

    @property
    def daily_production(self):
        return self.get_trend("DAY", "production")

    @property
    def daily_production_pct(self):
        return self.get_trend("DAY", "production_pct")

    @property
    def daily_net_production(self):
        return self.get_trend("DAY", "net_production")

    @property
    def daily_from_grid(self):
        return self.get_trend("DAY", "from_grid")

    @property
    def daily_to_grid(self):
        return self.get_trend("DAY", "to_grid")

    @property
    def daily_solar_powered(self):
        return self.get_trend("DAY", "solar_powered")

    @property
    def weekly_usage(self):
        return self.get_trend("WEEK", "consumption")

    @property
    def weekly_production(self):
        return self.get_trend("WEEK", "production")

    @property
    def weekly_production_pct(self):
        return self.get_trend("WEEK", "production_pct")

    @property
    def weekly_net_production(self):
        return self.get_trend("WEEK", "net_production")

    @property
    def weekly_from_grid(self):
        return self.get_trend("WEEK", "from_grid")

    @property
    def weekly_to_grid(self):
        return self.get_trend("WEEK", "to_grid")

    @property
    def weekly_solar_powered(self):
        return self.get_trend("WEEK", "solar_powered")

    @property
    def monthly_usage(self):
        return self.get_trend("MONTH", "consumption")

    @property
    def monthly_production(self):
        return self.get_trend("MONTH", "production")

    @property
    def monthly_production_pct(self):
        return self.get_trend("MONTH", "production_pct")

    @property
    def monthly_net_production(self):
        return self.get_trend("MONTH", "net_production")

    @property
    def monthly_from_grid(self):
        return self.get_trend("MONTH", "from_grid")

    @property
    def monthly_to_grid(self):
        return self.get_trend("MONTH", "to_grid")

    @property
    def monthly_solar_powered(self):
        return self.get_trend("MONTH", "solar_powered")

    @property
    def yearly_usage(self):
        return self.get_trend("YEAR", "consumption")

    @property
    def yearly_production(self):
        return self.get_trend("YEAR", "production")

    @property
    def yearly_production_pct(self):
        return self.get_trend("YEAR", "production_pct")

    @property
    def yearly_net_production(self):
        return self.get_trend("YEAR", "net_production")

    @property
    def yearly_from_grid(self):
        return self.get_trend("YEAR", "from_grid")

    @property
    def yearly_to_grid(self):
        return self.get_trend("YEAR", "to_grid")

    @property
    def yearly_solar_powered(self):
        return self.get_trend("YEAR", "solar_powered")

    @property
    def active_devices(self):
        return [d["name"] for d in self._realtime.get("devices", {})]

    @property
    def time_zone(self):
        return self._monitor.get("time_zone", "")

    def trend_start(self, scale):
        """Return start of trend last updated."""
        if "start" not in self._trend_data[scale]:
            return None
        start_iso = self._trend_data[scale]["start"].replace("Z", "+00:00")
        return datetime.strptime(start_iso, "%Y-%m-%dT%H:%M:%S.%f%z")

    def get_trend(self, scale, key):
        """Return trend data item from last update."""
        if isinstance(key, bool):
            key = "production" if key is True else "consumption"
        else:
            key = "consumption" if key == "usage" else key
        if key not in self._trend_data[scale]:
            return 0
        # Perform a check for a valid type
        if not isinstance(self._trend_data[scale][key], (dict, float, int)):
            return 0
        if isinstance(self._trend_data[scale][key], dict):
            total = self._trend_data[scale][key].get("total", 0)
        else:
            total = self._trend_data[scale][key]
        return total



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

version = '0.12.0'
setup(
    name = 'sense_energy',
    packages = ['sense_energy'], 
    install_requires=[
        'async_timeout>=3.0.1',
        'orjson',
        'requests',
        'websocket-client',
        'websockets',
        'aiohttp',
        "kasa-crypt>=0.2.0",
    ], 
    version = version,
    description = 'API for the Sense Energy Monitor',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author = 'scottbonline',
    author_email = 'scottbonline@gmail.com',
    url = 'https://github.com/scottbonline/sense',
    download_url = 'https://github.com/scottbonline/sense/archive/'+version+'.tar.gz',
    keywords = ['sense', 'energy', 'api'], 
    classifiers = [
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)
